/*-------------------------------------------------------------------*/
/* PROCESS.C      Pasi Frnti                                        */
/*                                                                   */
/* ---------------------------------								 */
/* Generic image processing software								 */
/* ---------------------------------								 */
/* Resolution:                                                       */
/*     - Reduction (subsampling,averaging)                           */
/*     - Enlargement (copy,interpolation)                            */
/* Histrogram:                                                       */
/*    - Addition / negation                                          */
/*    - Range compression                                            */
/*    - Histogram equalization                                       */
/*    - Quantization                                                 */
/* Pixel operations:                                                 */
/*    - Noise generation                                             */
/*    - Pixel transition (one pixel right)                           */
/* Spatial operations:                                               */
/*    - Smoothing                                                    */
/*    - Sharpening                                                   */
/*    - Edge detection                                               */
/*    - Variance transform                                           */
/* Color conversions:                                                */
/*    - Color separation                                             */
/*    - Bit slicing                                                  */
/*    - Pseudo coloring                                              */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "PROCESS"
#define VersionNumber  "Version 0.04"
#define LastUpdated    "12.4.99"

/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <values.h>
#include <limits.h>
#include <math.h>
 
#define  FACTFILE  "process.fac"

#include "parametr.c"
#include "pgm.h"
#include "dip.h"
#include "image.h"
#include "file.h"
#include "memctrl.h"
#include "owntypes.h"
#include "random.h"

/* ------------------------ Global variables ---------------------------- */

int  Mapping[3][256];
int  ActiveColor[3];
int  InputMid;
int  InputMax;
int  OutputMax;


/*-----------------------------  M i s c  ---------------------------*/

#if defined(max) || defined(min)
#undef max
#undef min
#endif
#define  max(a,b) ((a) > (b) ? (a) : (b))
#define  min(a,b) ((a) < (b) ? (a) : (b))
#define  ScaleBetween(a,b,c)  max(a, min(b, c))
#define  log2(value)		  (log(value) / log(2))
#define  round(a)             ((int) (a+0.5))


/* =================  T I M E   F U N C T I O N S  ==================*/


static long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


static void PrintTime(long start)
{
  time_t  tmp;
  int     hours, minutes, seconds;
  long    elapsed;

  elapsed  = (long)time(&tmp) - start;
  hours    = (int)(elapsed / 3600);
  elapsed  = (int)(elapsed % 3600);
  minutes  = (int)(elapsed / 60);
  seconds  = (int)(elapsed % 60);

  switch( Value(QuietLevel) )
    {
    case 0:  break;
    case 1:  printf(" %i:%02i:%02i", hours, minutes, seconds);
             break;
    default: printf("Time: %i:%02i:%02i\n", hours, minutes, seconds);
             break;
    }
}


/*==========================  P R I N T  ============================*/


void PrintInfo(void)
{
  printf("%s     %s                (last modified %s)\n",ProgName,VersionNumber,LastUpdated);
  printf("\nDigital image processing software for testing purposes.\n");
  printf("Usage: %s inputfile outputfile [%coptions]\n", ProgName, OPTION_SYMBOL);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s     %s\n",ProgName,VersionNumber);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* InputFileName, char* OutputFileName,
                               IMAGE* InputImage, IMAGE* OutputImage)
{
  if( Value(QuietLevel) >= 1 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Input file:       %12s  (%ix%i)\n", InputFileName,
            InputImage->ImageSizeX, InputImage->ImageSizeY);
    printf("Output file:      %12s  (%ix%i)\n", OutputFileName,
            OutputImage->ImageSizeX, OutputImage->ImageSizeY);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


static void PrintProgress(int y, int total)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("%5.1f %% done.\r", (float)(y + 1) * 100 / total);
    fflush(stdout);
    }
}


/* ================ File type and operative components ============= */
/*-------------------------------------------------------------------*/
/*                   IN    Process  OUT   Other:                     */
/*                   ---   -------  ---   ----------                 */
/* Hist.Equal.:      any   chosen   same  Two passes needed          */
/* Filtering:        any   chosen   same                             */
/* ColorSeparation:  RGB   joint     G                               */
/* PseudoColoring:    G    joint    RGB   not with color separation  */
/*-------------------------------------------------------------------*/


static void InitializeImageParameters(IMAGE* InputImage, IMAGE* OutputImage)
{
  int v;

  if( Value(ColorSeparation) && InputImage->FileType!=PPM )
     {
     printf("ERROR: Input image type must be PPM for COLOR SEPARATION.\n");
     exit(-1);
     }
  if( Value(PseudoColoring) && InputImage->FileType==PPM )
     {
     printf("ERROR: Input image type is already PPM. Cannot apply PSEUDO COLORING.\n");
     exit(-1);
     }
  if( Value(ColorSeparation) && Value(PseudoColoring) )
     {
     printf("ERROR: Cannot select both COLOR SEPARATION and PSEUDO COLORING.\n");
     exit(-1);
     }

  v = Value(TargetColor)==ALL ? YES : NO;
  ActiveColor[0] = ActiveColor[1] = ActiveColor[2] = v;
  ActiveColor[0] |= Value(TargetColor)==ONLY_RED;
  ActiveColor[1] |= Value(TargetColor)==ONLY_GREEN;
  ActiveColor[2] |= Value(TargetColor)==ONLY_BLUE;

  OutputImage->FileType = InputImage->FileType;
  if(Value(Thresholding))    OutputImage->FileType = PBM;
  if(Value(ColorSeparation)) OutputImage->FileType = PGM;
  if(Value(PseudoColoring))  OutputImage->FileType = PPM;

  InputMid  = InputImage->FileType==PBM  ? 1 : 128;
  InputMax  = InputImage->FileType==PBM  ? 1 : 255;
  OutputMax = OutputImage->FileType==PBM ? 1 : 255;
}


/* ======================== Global transforms ====================== */


static int StretchValue(int v)
{
  int   a,b,high,low,range;
  float k1,k2,r;

  low = Value(StretchLow);
  high = Value(StretchHigh);
  range = high - low;
  k1 = Value(StretchFactor);
  k2 = (float) (255 - range*k1) / (255-range);
  a = k2*low;
  b = a + k1*range;
  if(v<low)       r = k2*v;
  else if(v<high) r = a+k1*(v-low);
  else            r = b+k2*(v-high);
  return( round(r) );
}


/*-------------------------------------------------------------------*/


static void PreprocessImage(IMAGE* Image)
{
  long  Histogram[3][256];
  int   i,v,color;

  for(color=0; color<ColorBands(Image); color++)
     for(i=0; i<=Image->MaxValue; i++)
        Mapping[color][i] = i;

  if(Value(HistogramEqualization))
     {
     CalculateHistogram(Image, Histogram);
     EqualizeHistogram(Image, Histogram, Mapping);
     }

  for(color=0; color<ColorBands(Image); color++)
     for(i=0; i<=Image->MaxValue; i++)
       {
       v = Mapping[color][i];
       if(Value(Addition))
          v = ScaleBetween(0, v+Value(Addition), OutputMax);
       if(Value(ContrastStretching))
          v = ScaleBetween(0, StretchValue(v), OutputMax);
	   if(Value(Negation))
          v = 255 - v;
       if(Value(RangeCompression))
          v = ScaleBetween(0, Value(RangeCompression)*log2(v+1), OutputMax);
       if(Value(Thresholding))
          v = ThresholdPixel(v,Value(ThresholdValue));
       if(Value(Quantization)<8)
          v = QuantizePixel(v,Value(Quantization));
       Mapping[color][i] = ScaleBetween(0, v, OutputMax);
       if( Value(QuietLevel)>=3 ) printf("Map[%i,%i] = %i)\n",color,i,Mapping[color][i]);
	   }
}


/* =========================  P R O C E S S  ======================= */


static void ShiftLine(IMAGE* Image)
{
  int  x;
  BYTE r,g,b;

  for(x=Image->ImageSizeX; x>=2; x--)
	 {
	 GetBufferPnmPixel(Image, x-1, 0, &r, &g, &b);
	 PutBufferPnmPixel(Image, x, 0, r, g, b);
	 }
}


/*-------------------------------------------------------------------*/
#define  Wsum  MaskOperation


int  FilterPixel(IMAGE* Image, int x, int y, int color)
{
  int	mean,result;
  BYTE	v;
  float sd;

  switch( Value(Filtering) )
    {
	case SMOOTH:	result = (int) (Wsum(Image,x,y,color,1) / 9.0); break;
	case SHARP: 	result = (int) Wsum(Image,x,y,color,2); break;
    case MEDIAN:    result = (int) BlockMedian(Image,x,y,color); break;
	case SOBEL_XY:	result = abs( (int) Wsum(Image,x,y,color,3) ) +
                             abs( (int) Wsum(Image,x,y,color,4) ); break;
	case SOBEL_X:	result = abs( (int) Wsum(Image,x,y,color,3) ); break;
	case SOBEL_Y:	result = abs( (int) Wsum(Image,x,y,color,4) ); break;
    case VARIANCE:  result = (int) BlockVariance(Image,x,y,color); break;
    case SD:        sd     = sqrt(BlockVariance(Image,x,y,color));
                    result = (int) (4*sd); break;
    case LOCAL_BIN: mean   = (int) BlockMean(Image,x,y,color);
                    v      = GetBufferPixel(Image,x,y,color);
                    result = (v>=mean) ? OutputMax : 0; break;
	default:		result = 0; break;
    }
  result = ScaleBetween(0, result, OutputMax);
  return( result );
}


/*-------------------------------------------------------------------*/


static int ProcessPixel(IMAGE* Image, int x, int y, int c, int v)
{
  #define NoiseProp  (Value(NoiseParameter) / 100.0)
  int r=v;

  if(Value(Filtering))              r = FilterPixel(Image, x, y, c);
  if(Value(NoiseType)==SALT_PEPPER) r = GenerateSaltPepperNoise(r,NoiseProp);
  if(Value(NoiseType)==ADDITIVE)    r = GenerateAdditiveNoise(r,NoiseProp);
  return(r);
}


/*-------------------------------------------------------------------*/


static void  ProcessImage(IMAGE* InputImage, IMAGE* OutputImage)
{
  int  x,y;
  BYTE r1,g1,b1;
  int  r2,g2,b2;

  for(y=1; y<=InputImage->ImageSizeY; y++)
    {
    ScrollBuffer(InputImage);
    for(x=1; x<=InputImage->ImageSizeX; x++)
      {
      GetBufferPnmPixel(InputImage, x, 0, &r1, &g1, &b1);
      r2 = Mapping[0][r1];
      g2 = Mapping[1][g1];
      b2 = Mapping[2][b1];
      if(ActiveColor[0]) r2 = ProcessPixel(InputImage,x,0,0,r2);
      if(ActiveColor[1]) g2 = ProcessPixel(InputImage,x,0,1,g2);
      if(ActiveColor[2]) b2 = ProcessPixel(InputImage,x,0,2,b2);
      if(Value(ColorSeparation)) r2 = SeparateColor(r2, g2, b2, Value(ColorSeparation));
	  if(Value(PseudoColoring)) CalculatePseudoColor(r2, &r2, &g2, &b2, Value(PseudoColoring));
      r2 = (BYTE) ScaleBetween(0, r2, OutputMax);
      g2 = (BYTE) ScaleBetween(0, g2, OutputMax);
      b2 = (BYTE) ScaleBetween(0, b2, OutputMax);
      PutBufferPnmPixel(OutputImage, x, 0, r2, g2, b2);
      }
    if(Value(MoveImage)) ShiftLine(OutputImage);
    WriteCurrentLine(OutputImage);
    PrintProgress(y, InputImage->ImageSizeY);
    }
}


/*===========================  M A I N  =============================*/


void main(int argc, char* argv[])
{
  char                InputFileName[64] = { '\0' };
  char                OutputFileName[64] = { '\0' };
  FILE*               InputFile;
  FILE*               OutputFile;
  IMAGE               InputImage;
  IMAGE               OutputImage;
  FILETYPE            ft;
  long                watch;
  ParameterInfo paraminfo[2] = { { InputFileName,  "", 0, INFILE  },
                                 { OutputFileName, "", 0, OUTFILE } };

  initrandom(0);
  ParseParameters(argc, argv, 2, paraminfo);

  ft = DetermineFileFormat(InputFileName);
  CheckFileName(InputFileName, FileFormatName(ft));
  InputFile  = OpenFile(InputFileName,	INPUT,	NO);
  InitializeImage(InputFile,  &InputImage,  INPUT,  WINDOWBUFFER, 3);

  CopyImageHeader(&InputImage, &OutputImage);
  InitializeImageParameters(&InputImage, &OutputImage);
  CheckFileName(OutputFileName, FileFormatName(OutputImage.FileType));
  OutputFile = OpenFile(OutputFileName, OUTPUT, Value(OverWrite));
  InitializeImage(OutputFile, &OutputImage, OUTPUT, WINDOWBUFFER, 1);
  PrintOperatingInfo(InputFileName, OutputFileName, &InputImage, &OutputImage);
  SetWatch(&watch);

  PreprocessImage(&InputImage);
  ProcessImage(&InputImage, &OutputImage);

  PrintTime(watch);
  FreeImage(&InputImage);
  FreeImage(&OutputImage);
  fclose(InputFile);
  fclose(OutputFile);
  checkmemory();
}

